/*-------------------------------------------------------------------
 * Vortaro als Konsolenprogramm
 */
#include	<stdio.h>
#include	<string.h>
#include	<strings.h>
#include	<stdlib.h>
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<unistd.h>
#include	<fcntl.h>

#define	FALSE	0
#define	TRUE	-1

#define	RET_OK		0
#define	RET_NOMEM	1
#define	RET_NOFILE	2

/*
 * Fuer die Wortliste der Rueckuebersetzung
 */
struct wl {
  struct wl	*next;
  char		wort[1000];
};

/*-------------------------------------------------------------------
 * Global verfuegbare Variablen
 */
char	*eo, *de;
char	vorne[1000], hinten[1000];

/*-------------------------------------------------------------------
 * Fuer die Konvertierung der Zeichen gilt:
 *
 * cx : c4 89  -  196 137
 * Cx : c4 88  -  196 136
 *
 * gx : c4 9d  -  196 157
 * Gx : c4 9c  -  196 156
 *
 * hx : c4 a5  -  196 165
 * Hx : c4 a4  -  196 164
 *
 * jx : c4 b5  -  196 181
 * Jx : c4 b4  -  196 180
 *
 * sx : c5 9d  -  197 157
 * Sx : c5 9c  -  197 156
 *
 * ux : c5 ad  -  197 173
 * Ux : c5 ac  -  197 172
 */

/*===================================================================
 * Zunaechst kommen die (beiden) Routinen fuer die Uebersetzung von
 * E(speranto)->X(-Schreibweise).
 * Sie bedienen sich einer gemeinsamen Ersetzungsroutine, die schlicht
 * im String zwei Zeichenpaare austauscht.
 *-------------------------------------------------------------------
 * Ersetzen von zwei Buchstaben durch die entsprechende Bytefolge
 */
void ersetze(char *worin, char *was, int eins, int zwei){
  char	*cptr;

  while (NULL != (cptr = strstr(worin, was))){
    *cptr = (char)eins;
    cptr++;
    *cptr = (char)zwei;
  }
}

/*-------------------------------------------------------------------
 * In dem uebergebenen String muessen alle Esperanto-Zeichen in ihre
 * x-Schreibweise gebracht werden.
 */
void e2x(char *zeile){
  char	*chef, *horst;

  chef = zeile;

  while ('\0' != *chef){
    horst = chef;
    horst++;
    if ((char)196 == *chef){
      if ((char)137 == *horst){*chef = 'c'; *horst = 'x';}
      if ((char)136 == *horst){*chef = 'C'; *horst = 'x';}
      if ((char)157 == *horst){*chef = 'g'; *horst = 'x';}
      if ((char)156 == *horst){*chef = 'G'; *horst = 'x';}
      if ((char)165 == *horst){*chef = 'h'; *horst = 'x';}
      if ((char)164 == *horst){*chef = 'H'; *horst = 'x';}
      if ((char)181 == *horst){*chef = 'j'; *horst = 'x';}
      if ((char)180 == *horst){*chef = 'J'; *horst = 'x';}
    }
    if ((char)197 == *chef){
      if ((char)157 == *horst){*chef = 's'; *horst = 'x';}
      if ((char)156 == *horst){*chef = 'S'; *horst = 'x';}
      if ((char)173 == *horst){*chef = 'u'; *horst = 'x';}
      if ((char)172 == *horst){*chef = 'U'; *horst = 'x';}
    }
    chef++;
  }
}

/*-------------------------------------------------------------------
 * Ausgabe eines Strings in Esperanto-Zeichen
 */
void x2e(char *eo){
  /*
   * Die Ausgabe eines Esperantostrings ist eigentlich recht einfach.
   * Es muessen nur alle Xx-Kombinationen durch den entsprechenden
   * Bytekode ersetzt werden
   */
  ersetze(eo, "cx", 196, 137);
  ersetze(eo, "Cx", 196, 136);
  ersetze(eo, "gx", 196, 157);
  ersetze(eo, "Gx", 196, 156);
  ersetze(eo, "hx", 196, 165);
  ersetze(eo, "Hx", 196, 164);
  ersetze(eo, "jx", 196, 181);
  ersetze(eo, "Jx", 196, 180);
  ersetze(eo, "sx", 197, 157);
  ersetze(eo, "Sx", 197, 156);
  ersetze(eo, "ux", 197, 173);
  ersetze(eo, "Ux", 197, 172);
}

/*===================================================================
 * Hier ein paar Hilfsroutinen
 *
 * Loeschen aller fuehrenden und anhaengenden Leerzeichen
 */
void trim(char *zeile){
  char	hier[1000], *sptr, *eptr;

  strcpy(hier, zeile);
  sptr = hier;
  while ((' ' == *sptr) ||
	 ('\t' == *sptr) ||
	 ('\n' == *sptr)) sptr++;
  
  eptr = &hier[strlen(hier) - 1];
  while ((' ' == *eptr) ||
	 ('\t' == *eptr) ||
	 ('\n' == *eptr)) eptr--;
  eptr++;
  
  *eptr = '\0';
  strcpy(zeile, sptr);
}

/*-------------------------------------------------------------------
 * Loeschung eines Klammerausdrucks aus einer Zeile. Die oeffnende und
 * die schliessende Klammer werden angegeben
 */
void loesche_klammer(char *zeile, char ka, char kz){
  char	mit[1000], ohne[1000], *sptr, *dptr, inklammer;

  strcpy(mit, zeile);
  ohne[0] = '\0';
  inklammer = FALSE;

  sptr = mit;
  dptr = ohne;

  while ('\0' != *sptr){
    if (ka == *sptr) inklammer = TRUE;

    if (FALSE == inklammer){
      *dptr = *sptr;
      dptr++;
    }

    if (kz == *sptr) inklammer = FALSE;

    sptr++;
  }
  *dptr = '\0';
  strcpy(zeile, ohne);
}

void keineklammern(char *zeile){
  char	vor[1000];

  strcpy(vor, zeile);
  loesche_klammer(vor, '(', ')');
  /*
  loesche_klammer(vor, '[', ']');
  loesche_klammer(vor, '{', '}');
  */
  strcpy(zeile, vor);
}

/*===================================================================
 * Nun kommt die Routine, die stumpf die Woerterbuecher einliest. Das
 * Esperanto-Woerterbuch wird an die Variable 'eo' gehaengt und das
 * deutsche Woerterbuch an die Variable 'de'
 *-------------------------------------------------------------------
 * Einlesen der Woerterbuecher in den Speicher
 */
int read_vortarojn(void){
  int		ergebnis;
  struct stat	status;
  char		dname[1000];
  int		datei;

  ergebnis = RET_OK;

  strcpy(dname, "esperanto.txt");
  stat(dname, &status);
  printf("Größe der Esperantodatei / Egeco de la Esperanto-dosiero: %d Bytes\n", (int)status.st_size);

  eo = (char *)malloc((int)status.st_size + 1);
  if (NULL != eo){
    datei = open(dname, O_RDONLY);
    if (-1 != datei){
      read(datei, (void *)eo, status.st_size);
      close(datei);
      printf("Datei gelesen / Dosiero legita\n");
      eo[(int)status.st_size] = '\0';
    }
    else{
      printf("Fehler beim Oeffnen der Datei\n");
      printf("Eraro ĉe malfermi la dosieron\n");
      ergebnis = RET_NOFILE;
    }
  }
  else{
    printf("Kein ausreichender Speicher\n");
    printf("Ne sufiĉa memoro\n");
    ergebnis = RET_NOMEM;
  }

  strcpy(dname, "deutsch.txt");
  stat(dname, &status);
  printf("Größe der Deutschdatei / Egeco de la germana dosiero: %d Bytes\n", (int)status.st_size);

  de = (char *)malloc((int)status.st_size + 1);
  if (NULL != de){
    datei = open(dname, O_RDONLY);
    if (-1 != datei){
      read(datei, (void *)de, status.st_size);
      close(datei);
      printf("Datei gelesen / Dosiero legita\n");
      de[(int)status.st_size] = '\0';
    }
    else{
      printf("Fehler beim Oeffnen der Datei\n");
      printf("Eraro ĉe malfermi la dosieron\n");
      ergebnis = RET_NOFILE;
    }
  }
  else{
    printf("Kein ausreichender Speicher\n");
    printf("Ne sufiĉa memoro\n");
    ergebnis = RET_NOMEM;
  }

  return(ergebnis);
}

/*===================================================================
 * So, nun kommen die Routinen, die mit den Zeilen des Woerterbuchs zu
 * tun haben. Das sind zum einen die, die einen Zeiger auf die
 * naechste oder die vorherige Zeile versetzen, aber dann auch die,
 * welche mit der Zerlegung der Zeilen zu tun haben.
 *-------------------------------------------------------------------
 * Finden der naechsten Zeile
 */
char *nz(char *aktuell){
  char	*ergebnis;

  ergebnis = aktuell;
  /*
   * Solange nicht das Ende einer Zeile oder das Ende aller Zeilen
   * erreicht ist, wird der Zeiger erhoeht
   */
  while (('\n' != *ergebnis) && ('\0' != *ergebnis))
    ergebnis++;

  /*
   * Gibt es noch eine weiter Zeile, dann wird auf deren Anfang
   * erhoeht
   */
  if ('\n' == *ergebnis) ergebnis++;

  /*
   * Wenn es keine weitere Zeile gibt, dann wird wieder die aktuelle
   * Zeile zurueck gegeben
   */
  if ('\0' == *ergebnis) ergebnis = aktuell;

  return(ergebnis);
}

/*-------------------------------------------------------------------
 * Finden der vorherigen Zeile
 */
char *vz(char *liste, char *aktuell){
  char	*ergebnis;

  ergebnis = aktuell;
  /*
   * Wenn der Zeiger nicht auf dem Anfang der ersten Zeile steht, dann
   * muss er erst auf das letzte Zeichen der vorherigen Zeile
   * erniedrigt werden
   */
  if (ergebnis != liste){ergebnis--; ergebnis--;}

  /*
   * Nun wird der Zeiger so lange erniedrigt, bis entweder das Ende
   * der vorherigen Zeile oder die erste Zeile erreicht wurde
   */
  while ((ergebnis != liste) && (*ergebnis != '\n'))
    ergebnis--;

  /*
   * Wenn nicht der Anfang aller Zeilen erreicht wurde, dann muss der
   * Zeiger wieder auf das erste Zeichen der folgenden Zeile erhoeht
   * werden
   */
  if ('\n' == *ergebnis) ergebnis++;

  return(ergebnis);
}

/*-------------------------------------------------------------------
 * Die folgende Routine teilt eine Zeile in Wort und Uebersetzung und
 * legt die Ergebnisse in den globalen Variablen 'vorne' und 'hinten'
 * ab.
 */
void teile_zeile(char *zeile){
  char	*vptr, *hptr, zwischen;

  vptr = hptr = zeile;
  while (';' != *vptr) vptr++;
  zwischen = *vptr;
  *vptr = '\0';
  strcpy(vorne, hptr);

  *vptr = zwischen;
  vptr++;
  hptr = vptr;

  while (('\0' != *vptr) && ('\n' != *vptr)) vptr++;
  zwischen = *vptr;
  *vptr = '\0';
  strcpy(hinten, hptr);
  *vptr = zwischen;
}

/*-------------------------------------------------------------------
 * Finden eines passenden Eintrags in einer Liste
 */
char *finde_eintrag(char *liste, char *muster){
  char	*ergebnis, *altezeile;
  int	laenge, vergleichswert;

  laenge = strlen(muster);
  ergebnis = liste;
  altezeile = NULL;
  do{
    vergleichswert = strncasecmp(muster, ergebnis, laenge);

    if (0 < vergleichswert){
      altezeile = ergebnis;
      ergebnis = nz(ergebnis);
      if (altezeile == ergebnis) vergleichswert = -1;
    }
  } while (0 < vergleichswert);

  return(ergebnis);
}

/*-------------------------------------------------------------------
 * Ausgabe einer Zeile aus einer Liste
 * Der ebenfalls uebergebene boolsche Operator hat den Sinn zwischen
 * einer vollstaendigen und einer unvollstaendigen zu unterscheiden.
 */
void zeilenausgabe(char *zeile, char vollstaendig){

  /*
   * Wort un Uebersetzung in den globalen Variablen ablegen
   */
  teile_zeile(zeile);

  x2e(vorne);
  x2e(hinten);

  printf("%s", vorne);
  if (TRUE == vollstaendig) printf(" -> %s", hinten);
  printf("\n");
}

/*-------------------------------------------------------------------
 * Die folgende Routine erledigt die Rueckuebersetzung der
 * Esperanto-Worter, die in einer Definition eines deutschen Wortes
 * steht.
 */
void rueckuebersetzung(char *deutscheswort){
  struct wl	*woerter, *wptr;
  char		*vptr, *hptr, zwischen;

  woerter = NULL;

  /*
   * Erst mal wird die Uebersetzung geholt
   */
  teile_zeile(deutscheswort);

  hptr = hinten;
  vptr = hinten;

  while (('\0' != *vptr) && ('\n' != *vptr)){
    while (('\0' != *vptr) && ('\n' != *vptr) && (',' != *vptr)) vptr++;
    /*
     * jetzt steht vptr am Ende des Wortes und hptr am Anfang. Dieses
     * Wort kann nun in die Wortliste aufgenommen werden.
     */
    zwischen = *vptr;
    *vptr = '\0';
    if (NULL == woerter){
      woerter = (struct wl *)malloc(sizeof(struct wl));
      wptr = woerter;
    }
    else{
      wptr = woerter;
      while (NULL != wptr->next) wptr = wptr->next;
      wptr->next = (struct wl *)malloc(sizeof(struct wl));
      wptr = wptr->next;
    }
    wptr->next = NULL;
    strcpy(wptr->wort, hptr);

    *vptr = zwischen;
    if (('\0' != *vptr) && ('\n' != *vptr)){
      vptr++;
      hptr = vptr;
    }
  }

  /*
   * Nun stehen die Woerter in der Wortliste, allerdings noch in roher
   * Form. Diese muss nun noch bereinigt werden.
   */

  /*
   * Zum Schluss wird der Speicher wieder frei gegeben und dabei fuer
   * jedes der uebrig gebliebenen Woerter die volle Zeile ausgegeben
   */
  while (NULL != woerter){
    keineklammern(woerter->wort);
    trim(woerter->wort);

    if (NULL == strchr(woerter->wort, ' ')){
      vptr = finde_eintrag(eo, woerter->wort);
      if (NULL != vptr) zeilenausgabe(vptr, TRUE);
    }
    wptr = woerter->next;
    free(woerter);
    woerter = wptr;
  }
}

/*-------------------------------------------------------------------
 * Hier kommt die oberste Ausgaberoutine. Sie bekommt einen Zeiger auf
 * die Zeile mit einem Esperantowort und einen auf eine Zeile mit
 * einem deutschen Wort.
 * Diese Zeilen muessen ausgegeben werden und dann noch die
 * Rueckuebersetzung der Esperanto-Uebersetzungen
 */
void ausgabe(char *esperanto, char *deutsch){
  char	*zeige;

  /*
   * Erst kommt der Eintrag Esperanto -> Deutsch
   * Es werden auch zwei Zeilen davor und danach ausgegeben, die
   * allerdings ohne Uebersetzung
   */
  printf("#---------------------\n");
  zeige = esperanto;
  zeige = vz(eo, zeige);
  zeige = vz(eo, zeige);
  zeilenausgabe(zeige, FALSE);
  zeige = nz(zeige);
  zeilenausgabe(zeige, FALSE);
  zeilenausgabe(esperanto, TRUE);
  zeige = nz(esperanto);
  zeilenausgabe(zeige, FALSE);
  zeige = nz(zeige);
  zeilenausgabe(zeige, FALSE);

  /*
   * Und nun dieselbe Nummer fuer den Eintrag Deutsch -> Esperanto
   */
  printf("#---------------------\n");
  zeige = deutsch;
  zeige = vz(de, zeige);
  zeige = vz(de, zeige);
  zeilenausgabe(zeige, FALSE);
  zeige = nz(zeige);
  zeilenausgabe(zeige, FALSE);
  zeilenausgabe(deutsch, TRUE);
  zeige = nz(deutsch);
  zeilenausgabe(zeige, FALSE);
  zeige = nz(zeige);
  zeilenausgabe(zeige, FALSE);

  /*
   * Und nun die Eintraege aller Esperanto-Woerter, die als
   * Uebersetzungsvorschlag fuer ein deutsches Wort angeboten werden.
   * Da dieser Vorgang recht komplex ist, wird er in eine eigene
   * Routine gepackt
   */
  printf("#---------------------\n");
  rueckuebersetzung(deutsch);
}

/*-------------------------------------------------------------------
 * Hauptprogramm
 */
int main(int argc, char *argv[]){
  int		ergebnis;
  char		eingabe[1000], ende;
  char		*eptr, *dptr;

  /*
   * Einlesen der Dateien mit den Woerterbuechern
   */
  ergebnis = read_vortarojn();

  /*
   * Hier beginnt nun der Hauptloop.
   * Es wird ein Wort eingegeben. Danach werden die Sonderfaelle
   * getestet. Liegt kein solcher Sonderfall vor, dann wird versucht
   * das Wort zu uebersetzen und die Uebersetzung anzuzeigen. Danach
   * wird dann das naechste Wort eingelesen
   *
   * Sonderfaelle sind:
   * 'q' : Beenden des Loops
   * '-' : Ein Wort in der Liste zurueck
   * '+' : Ein Wort in der Liste weiter
   */
  if (RET_OK == ergebnis){
    if (1 == argc){
      ende = FALSE;
      eingabe[0] = '\0';
      do{
        if ('\n' != eingabe[0]){
	  printf("-> "); scanf("%s", eingabe);
	  e2x(eingabe);

	  if (1 == strlen(eingabe)){
	    switch (eingabe[0]){
	      case 'q' : {
	        ende = TRUE;
	        break;
	      }
	      case '-' : {
	        eptr = vz(eo, eptr);
	        dptr = vz(de, dptr);
	        break;
	      }
	      case '+' : {
	        eptr = nz(eptr);
	        dptr = nz(dptr);
	        break;
	      }
	      default : {
	        eptr = finde_eintrag(eo, eingabe);
	        dptr = finde_eintrag(de, eingabe);
	      }
	    }
	  }
	  else{
	    eptr = finde_eintrag(eo, eingabe);
	    dptr = finde_eintrag(de, eingabe);
	  }
	  /*
	   * Nun sollte der dptr und der eptr auf dem jeweils passenden
	   * Eintrag stehen, der dann auch ausgegeben werden kann
	   */
	  if (!ende){
            printf("\n\n>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>\n\n");
	    printf("Eingabe / Eingo: %s\n", eingabe);
	    ausgabe(eptr, dptr);
	  }
        }
      } while (!ende);
    }
    else{
      /*
       * Wenn ein Argument angegeben wurde, dann wird versucht dieses
       * zu uebersetzen
       */
      strcpy(eingabe, argv[1]);

      /*
       * Zur besseren Nutzbarkeit dieses Programms, wird es heute, am
       * 8.1.2019 verbessert.
       * Die eingegebene Zeile wird in jedem Fall in die x-Form der
       * Esperanto-Schreibweise gebracht.
       */
      e2x(eingabe);

      printf("\n>>>>>>>>>>>>>>>>>>\n %s\n", eingabe);
      eptr = finde_eintrag(eo, eingabe);
      dptr = finde_eintrag(de, eingabe);
      ausgabe(eptr, dptr);
    }
  }

  /*
   * Freigeben des Speichers
   */
  if (RET_NOMEM != ergebnis){
    free(eo);
    free(de);
  }
  return(ergebnis);
}

